from .bases.processor import Processor
from .bases.processor_interface import ProcessorInterface
from .mixins.pipeline_factory_interface_mixin import PipelineFactoryInterfaceMixin
from .mixins.pipeline_factory_mixin import PipelineFactoryMixin
from .mixins.pipeline_interface_mixin import PipelineInterfaceMixin
from .mixins.pipeline_mixin import PipelineMixin

__all__ = (
    "PipelineMixin",
    "PipelineInterfaceMixin",
    "PipelineFactoryMixin",
    "PipelineFactoryInterfaceMixin",
    "Processor",
    "ProcessorInterface",
)



from typing import Protocol, TypeVar

# extends: outside base
from ..._01_with_base import PipelineInterface as PipelineBaseInterface

# extends: outside mixins
from ..._02_with_pipeline_as_list import (
    PipelineInterfaceMixin as _02_WithPipelineAsListInterface,
)


class PipelineInterface(
    _02_WithPipelineAsListInterface,
    PipelineBaseInterface,
    Protocol,
):
    pass


TPipeline = TypeVar("TPipeline", bound=PipelineInterface, infer_variance=True)



from typing import Any, Optional, Protocol, Self

# extends: 3rd-party interface
from ....support.act_as_list import ActAsListInterface

# extends: self-bridge
from ..bridges.pipeline_interface import PipelineInterface

# uses: bridge interface
from ..bridges.stage_callable import (
    StageInstance,
    StageInstanceCollection,
)


class PipelineInterfaceMixin(
    ActAsListInterface[StageInstance, StageInstanceCollection],
    PipelineInterface,
    Protocol,
):
    def __init__(
        self,
        stage_instances: Optional[StageInstanceCollection] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def pipe(self, stage: StageInstance) -> Self: ...



# Re-exporting symbols
from .chained_processor import ChainedPipeline as ChainedPipeline
from .chained_processor import ChainedProcessor as ChainedProcessor
from .chained_processor import __all__ as _chained_processor_all
from .interruptible_processor import InterruptiblePipeline as InterruptiblePipeline
from .interruptible_processor import InterruptibleProcessor as InterruptibleProcessor
from .interruptible_processor import __all__ as _interruptible_processor_all

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    *_chained_processor_all,
    *_interruptible_processor_all,
)



import inspect
from typing import Any, Awaitable, Callable, Protocol

from .has_call_async_interface import HasCallAsyncInterface, TCallableReturn


class HasCallAsync(HasCallAsyncInterface, Protocol):
    async def _call(
        self,
        callable: Callable[..., TCallableReturn | Awaitable[TCallableReturn]],
        *args: Any,
        **kwds: Any,
    ) -> TCallableReturn:
        result = callable(*args, **kwds)

        if inspect.isawaitable(result):
            return await result

        return result



from typing import Any, Awaitable, Callable, Protocol, TypeVar

TCallableReturn = TypeVar("TCallableReturn", infer_variance=True)


class HasCallAsyncInterface(Protocol):
    async def _call(
        self,
        callable: Callable[..., TCallableReturn | Awaitable[TCallableReturn]],
        *args: Any,
        **kwds: Any,
    ) -> TCallableReturn: ...



from ...core import ChainedProcessor as ChainedProcessorBase
from ...core import T_in, T_out


class ChainedProcessor(ChainedProcessorBase[T_in, T_out]):
    """Default processor. Sequentially processes data through multiple stages.

    Example:
        ```python
        # Create processor
        processor = ChainedProcessor[int]()

        # Stages
        result = await processor.process(
            payload=5,
            stages=(
                lambda payload: payload + 1,
                lambda payload: payload * 2,
            ),
        )

        # Assert result
        assert result == 12
        ```
    """



from abc import abstractmethod
from typing import Protocol, Self

from .types import TItem, TCollection
from .act_as_list_interface import ActAsListInterface


class ActAsList(
    ActAsListInterface[TItem, TCollection],
    Protocol[TItem, TCollection],
):
    @abstractmethod
    def _get_items(self) -> TCollection:
        pass

    @abstractmethod
    def _set_items(self, items: TCollection) -> Self:
        pass

    @abstractmethod
    def _add_item(self, item: TItem) -> Self:
        pass



# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.23.0"

# Re-exporting symbols
from .core import Pipeline as Pipeline
from .core import PipelineFactory as PipelineFactory
from .core import PipelineFactoryInterface as PipelineFactoryInterface
from .core import PipelineInterface as PipelineInterface
from .core import Processor as Processor
from .core import ProcessorInterface as ProcessorInterface
from .core import Stage as Stage
from .core import StageCallable as StageCallable
from .core import StageClassOrInstance as StageClassOrInstance
from .core import StageCollection as StageCollection
from .core import StageInstance as StageInstance
from .core import StageInstanceCollection as StageInstanceCollection
from .core import StageInterface as StageInterface
from .core import T_in as T_in
from .core import T_out as T_out
from .core import __all__ as _core_all
from .processors import ChainedPipeline as ChainedPipeline
from .processors import ChainedProcessor as ChainedProcessor
from .processors import InterruptiblePipeline as InterruptiblePipeline
from .processors import InterruptibleProcessor as InterruptibleProcessor
from .processors import __all__ as _processor_all

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    *_core_all,
    *_processor_all,
)



from abc import abstractmethod
from typing import Any

# extends: 3rd-party concrete
from ....support.clonable import Clonable
from ....support.has_call_async import HasCallAsync

# uses: bridge interface
from ..bridges.stage_callable import StageInstance, StageInstanceCollection
from ..bridges.types import T_in, T_out

# implements: self-interface
from .processor_interface import ProcessorInterface as ImplementsInterface


class Processor(
    HasCallAsync,
    Clonable,
    ImplementsInterface[T_in, T_out],
):
    def __init__(self, *args: Any, **kwds: Any) -> None:
        """Constructor."""
        pass

    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: StageInstanceCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Process the given payload through the provided stages.

        Args:
            payload (T_in): The input payload to process.
            stages (StageInstanceCollection): The collection of stages to process the payload through.
            *args (Any): Additional positional arguments.
            **kwds (Any): Additional keyword arguments.

        Returns:
            T_out: The processed output.
        """
        pass

    # HasCallAsync: include callable type and `payload` in the signature
    async def _call(
        self,
        callable: StageInstance,
        payload: T_in,
        *args: Any,
        **kwds: Any,
    ) -> Any:
        """
        Alias to `process` method.
        """
        return await super()._call(callable, payload, *args, **kwds)



"""A collection of processors and their pipelines"""

# Re-exporting symbols
from _api.processors import ChainedPipeline as ChainedPipeline
from _api.processors import ChainedProcessor as ChainedProcessor
from _api.processors import InterruptiblePipeline as InterruptiblePipeline
from _api.processors import InterruptibleProcessor as InterruptibleProcessor
from _api.processors import __all__ as _processors_all

# pyright: reportUnsupportedDunderAll=false
__all__ = (*_processors_all,)



from .bases.pipeline import Pipeline
from .bases.pipeline_interface import PipelineInterface, TPipeline
from .bases.stage import Stage
from .bases.stage_callable import (
    StageCallable,
    StageClassOrInstance,
    StageCollection,
    StageInstance,
    StageInstanceCollection,
)
from .bases.stage_interface import StageInterface
from .bases.types import T_in, T_out

__all__ = (
    "Pipeline",
    "PipelineInterface",
    "Stage",
    "StageInterface",
    "StageCallable",
    "StageClassOrInstance",
    "StageCollection",
    "StageInstance",
    "StageInstanceCollection",
    "T_in",
    "T_out",
    "TPipeline",
)



from typing import Any, Optional, Protocol, Self

# extends: self-bridge
from ..bridges.pipeline_interface import PipelineInterface

# uses: bridge interface
from ..bridges.stage_callable import StageCollection


class PipelineInterfaceMixin(
    PipelineInterface,
    Protocol,
):
    def __init__(
        self,
        stages: Optional[StageCollection] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def _should_instantiate_stages(self) -> bool: ...

    def _instantiate_stages(self) -> Self: ...

    def with_stages(self, stages: StageCollection) -> Self: ...



from typing import Protocol

# extends: self-bridge
from ..bridges.pipeline_interface import PipelineInterface

# uses: bridge interface
from ..bridges.types import T_in, T_out


class PipelineInterfaceMixin(
    PipelineInterface,
    Protocol[T_in, T_out],
):
    pass



from typing import Generic

# extends: outside base
from ...plugins._99_with_pipeline_default_processor import (
    ChainedProcessor as ChainedProcessorBase,
)

# uses: bridge interface
from ..types import T_in, T_out


class ChainedProcessor(
    ChainedProcessorBase[T_in, T_out],
    Generic[T_in, T_out],
):
    pass



from typing import Any, Generic, Optional, Self

# uses: local base
from ..bases.processor_interface import ProcessorInterface

# uses: bridge interface
from ..bridges.types import T_in, T_out

# implements: self-interface
from .pipeline_interface_mixin import PipelineInterfaceMixin as ImplementsInterface


class PipelineMixin(
    ImplementsInterface[T_in, T_out],
    Generic[T_in, T_out],
):
    processor_class: Optional[type[ProcessorInterface[T_in, T_out]]]
    processor_instance: Optional[ProcessorInterface[T_in, T_out]]

    def __init__(
        self,
        processor_class: Optional[type[ProcessorInterface]] = None,
        processor_instance: Optional[ProcessorInterface] = None,
        processor: Optional[type[ProcessorInterface] | ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not hasattr(self, "processor_class"):
            self.processor_class = self._get_default_processor_class()

        if not hasattr(self, "processor_instance"):
            self.processor_instance = None

        if processor_class:
            self.processor_class = processor_class

        if processor_instance:
            self.processor_instance = processor_instance

        if processor:
            cloned = self.with_processor(processor)

            self.processor_class = cloned.processor_class

            self.processor_instance = cloned.processor_instance

        if not self.processor_instance:
            self.processor_instance = self._make_processor()

    def with_processor(
        self, processor: type[ProcessorInterface] | ProcessorInterface
    ) -> Self:
        """
        Attachs a processor (class or instance) to the pipeline.
        """
        if isinstance(processor, type):
            return self.with_processor_class(processor)

        return self.with_processor_instance(processor)

    def with_processor_instance(self, processor_instance: ProcessorInterface) -> Self:
        """
        Attachs a processor instance to the pipeline.
        """
        return self.clone(
            {
                "processor_class": processor_instance.__class__,
                "processor_instance": processor_instance,
            }
        )

    def with_processor_class(self, processor_class: type[ProcessorInterface]) -> Self:
        """
        Attachs a processor class to the pipeline.
        """
        return self.clone({"processor_class": processor_class})

    def _make_processor(self) -> ProcessorInterface:
        if self.processor_class is None:
            raise ValueError("Processor class not set")

        return self.processor_class()

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out:
        """
        Process the given payload through the pipeline.
        """
        if self.processor_instance is None:
            raise ValueError("Processor not set")

        return await self.processor_instance.process(
            payload=payload, stages=self._get_items(), *args, **kwds
        )

    def _get_default_processor_class(self) -> Optional[type[ProcessorInterface]]:
        return None

    def get_processor_instance(self) -> Optional[ProcessorInterface]:
        return self.processor_instance

    def get_processor_class(self) -> Optional[type[ProcessorInterface]]:
        return self.processor_class



from typing import Any, Generic, Optional, Self

# uses: local base
from ..bases.processor_interface import ProcessorInterface

# uses: bridge interface
from ..bridges.types import T_in, T_out

# implements: self-interface
from .pipeline_factory_interface_mixin import (
    PipelineFactoryInterfaceMixin as ImplementsInterface,
)


class PipelineFactoryMixin(
    ImplementsInterface[T_in, T_out],
    Generic[T_in, T_out],
):
    processor_class: Optional[type[ProcessorInterface[T_in, T_out]]] = None
    processor_instance: Optional[ProcessorInterface[T_in, T_out]] = None

    def __init__(
        self,
        processor_class: Optional[type[ProcessorInterface]] = None,
        processor_instance: Optional[ProcessorInterface] = None,
        processor: Optional[type[ProcessorInterface] | ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not self.processor_class:
            self.processor_class = processor_class or None

        if not self.processor_instance:
            self.processor_instance = processor_instance or None

        if processor:
            self.with_processor(processor)

    def with_processor(
        self, processor: type[ProcessorInterface] | ProcessorInterface
    ) -> Self:
        """
        Attachs a processor (class or instance) to the pipeline factory.
        """
        if isinstance(processor, type):
            return self.with_processor_class(processor)

        return self.with_processor_instance(processor)

    def with_processor_instance(self, processor_instance: ProcessorInterface) -> Self:
        """
        Attachs a processor instance to the pipeline factory.
        """
        self.processor_class = processor_instance.__class__

        self.processor_instance = processor_instance

        return self

    def with_processor_class(self, processor_class: type[ProcessorInterface]) -> Self:
        """
        Attachs a processor class to the pipeline factory.
        """
        self.processor_class = processor_class

        return self

    # ActAsFactory
    def _definition(self) -> dict[str, Any]:
        return {
            **super()._definition(),  # type: ignore
            "processor_class": self.processor_class,
            "processor_instance": self.processor_instance,
        }



from typing import Protocol

# extends: outside base
from ..._05_with_pipeline_factory import (
    PipelineFactoryInterface as PipelineFactoryBaseInterface,
)
from ..bridges.pipeline_interface import PipelineInterface

# uses: bridge interface
from ..bridges.types import T_in, T_out


class PipelineFactoryInterface(
    PipelineFactoryBaseInterface[PipelineInterface],
    Protocol[T_in, T_out],
):
    pass



from typing import Protocol

# extends: outside base
from ..plugins._05_with_pipeline_factory import (
    PipelineFactoryInterface as PipelineFactoryBaseInterface,
)

# extends: outside mixins
from ..plugins._06_with_pipeline_processor import (
    PipelineFactoryInterfaceMixin as _06_WithPipelineProcessorInterface,
)
from ..plugins._99_with_pipeline_default_processor import (
    PipelineFactoryInterfaceMixin as _99_WithPipelineDefaultProcessorInterface,
)

# uses: bridge interface
from .pipeline_interface import PipelineInterface
from .types import T_in, T_out


class PipelineFactoryInterface(
    _99_WithPipelineDefaultProcessorInterface[T_in, T_out],
    _06_WithPipelineProcessorInterface[T_in, T_out],
    PipelineFactoryBaseInterface[PipelineInterface[T_in, T_out]],
    Protocol[T_in, T_out],
):
    pass



from typing import Protocol

# extends: outside base
from ..._06_with_pipeline_processor import (
    ProcessorInterface as ProcessorBaseInterface,
)

# uses: bridge interface
from .types import T_in, T_out


class ProcessorInterface(
    ProcessorBaseInterface,
    Protocol[T_in, T_out],
):
    pass



from typing import Generic

# extends: outside base
from ..plugins._05_with_pipeline_factory import PipelineFactory as PipelineFactoryBase

# extends: outside mixins
from ..plugins._06_with_pipeline_processor import (
    PipelineFactoryMixin as _06_WithPipelineProcessor,
)
from ..plugins._99_with_pipeline_default_processor import (
    PipelineFactoryMixin as _99_WithPipelineDefaultProcessor,
)

# implements: self-interface
from .pipeline_factory_interface import PipelineFactoryInterface as ImplementsInterface

# uses: bridge interface
from .pipeline_interface import PipelineInterface
from .types import T_in, T_out


class PipelineFactory(
    _99_WithPipelineDefaultProcessor[T_in, T_out],
    _06_WithPipelineProcessor[T_in, T_out],
    PipelineFactoryBase[PipelineInterface[T_in, T_out]],
    ImplementsInterface[T_in, T_out],
    Generic[T_in, T_out],
):
    pass



# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.26.0"

# Re-exporting symbols
from _api.core import Pipeline as Pipeline
from _api.core import PipelineFactory as PipelineFactory
from _api.core import PipelineFactoryInterface as PipelineFactoryInterface
from _api.core import PipelineInterface as PipelineInterface
from _api.core import Processor as Processor
from _api.core import ProcessorInterface as ProcessorInterface
from _api.core import Stage as Stage
from _api.core import StageCallable as StageCallable
from _api.core import StageInterface as StageInterface

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    "Pipeline",
    "PipelineFactory",
    "PipelineFactoryInterface",
    "PipelineInterface",
    "Processor",
    "ProcessorInterface",
    "Stage",
    "StageCallable",
    "StageInterface",
)



from typing import Any, Optional, Protocol, Self

# uses: local base
from ..bases.processor_interface import ProcessorInterface

# extends: self-bridge
from ..bridges.pipeline_factory_interface import PipelineFactoryInterface

# uses: bridge interface
from ..bridges.types import T_in, T_out


class PipelineFactoryInterfaceMixin(
    PipelineFactoryInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        processor_class: Optional[type[ProcessorInterface]] = None,
        processor_instance: Optional[ProcessorInterface] = None,
        processor: Optional[type[ProcessorInterface] | ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def with_processor(
        self, processor: type[ProcessorInterface] | ProcessorInterface
    ) -> Self: ...

    def with_processor_instance(
        self, processor_instance: ProcessorInterface
    ) -> Self: ...

    def with_processor_class(
        self, processor_class: type[ProcessorInterface]
    ) -> Self: ...

    # ActAsFactory
    def _definition(self) -> dict[str, Any]: ...



from typing import Any, Awaitable, Protocol

from .types import T_in, T_out


class StageCallable(
    Protocol[T_in, T_out],
):
    """Callable object used in the pipeline"""

    def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out | Awaitable[T_out]: ...


StageInstance = StageCallable
"""Stage object"""

StageInstanceCollection = tuple[StageInstance, ...]
"""Collection of Stage objects"""

StageClassOrInstance = StageInstance | type[StageInstance]
"""Stage class or object"""

StageCollection = tuple[StageClassOrInstance, ...]
"""Collection of Stage classes or objects"""



from typing import Generic

# uses: bridge interface
from ..bridges.types import T_in, T_out

# implements: self-interface
from .pipeline_factory_interface_mixin import (
    PipelineFactoryInterfaceMixin as ImplementsInterface,
)

# uses: mixin-bridge concrete
from .pipeline_mixin import PipelineMixin as Pipeline


class PipelineFactoryMixin(
    ImplementsInterface[T_in, T_out],
    Generic[T_in, T_out],
):
    def _get_default_pipeline_class(self) -> type[Pipeline[T_in, T_out]]:
        return Pipeline[T_in, T_out]



from abc import ABC

# extends: outside base
from ..plugins._06_with_pipeline_processor import Processor as ProcessorBase

# implements: self-interface
from .processor_interface import ProcessorInterface as ImplementsInterface

# uses: bridge interface
from .types import T_in, T_out


class Processor(
    ProcessorBase[T_in, T_out],
    ImplementsInterface[T_in, T_out],
    ABC,
):
    """Processor Class"""

    pass



from typing import Protocol

# extends: outside base
from ...plugins._99_with_pipeline_default_processor import (
    ChainedProcessorInterface as ChainedProcessorBaseInterface,
)

# uses: final interface
from ..types import T_in, T_out


class ChainedProcessorInterface(
    ChainedProcessorBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass



from typing import Protocol, TypeVar

# extends: outside base
from ..._01_with_base import PipelineInterface as PipelineBaseInterface

# extends: outside mixins
from ..._02_with_pipeline_as_list import (
    PipelineInterfaceMixin as _02_WithPipelineAsListInterface,
)
from ..._04_with_pipeline_declared_stages import (
    PipelineInterfaceMixin as _04_WithPipelineDeclaredStagesInterface,
)
from ..._06_with_pipeline_processor import (
    PipelineInterfaceMixin as _06_WithPipelineProcessorInterface,
)
from ..._08_with_pipeline_as_stage import (
    PipelineInterfaceMixin as _08_WithPipelineAsStageInterface,
)

# uses: bridge interface
from .types import T_in, T_out


class PipelineInterface(
    _08_WithPipelineAsStageInterface[T_in, T_out],
    _06_WithPipelineProcessorInterface[T_in, T_out],
    _04_WithPipelineDeclaredStagesInterface,
    _02_WithPipelineAsListInterface,
    PipelineBaseInterface,
    Protocol[T_in, T_out],
):
    pass


TPipeline = TypeVar("TPipeline", bound=PipelineInterface, infer_variance=True)



from typing import Generic

# extends: outside base
from ..._06_with_pipeline_processor import (
    Processor as ProcessorBase,
)

# implements: self-interface
from .processor_interface import ProcessorInterface as ImplementsInterface

# uses: bridge interface
from .types import T_in, T_out


class Processor(
    ProcessorBase[T_in, T_out],
    ImplementsInterface[T_in, T_out],
    Generic[T_in, T_out],
):
    pass



from .pipeline import Pipeline
from .pipeline_factory import PipelineFactory
from .pipeline_factory_interface import PipelineFactoryInterface
from .pipeline_interface import PipelineInterface
from .processor import Processor
from .processor_interface import ProcessorInterface
from .processors.chained_processor import ChainedProcessor
from .processors.chained_processor_interface import ChainedProcessorInterface
from .stage import Stage
from .stage_callable import (
    StageCallable,
    StageClassOrInstance,
    StageCollection,
    StageInstance,
    StageInstanceCollection,
)
from .stage_interface import StageInterface
from .types import T_in, T_out

__all__ = (
    "Pipeline",
    "PipelineInterface",
    "Stage",
    "StageInterface",
    "Processor",
    "ProcessorInterface",
    "PipelineFactory",
    "PipelineFactoryInterface",
    "T_in",
    "T_out",
    "StageCallable",
    "StageInstance",
    "StageInstanceCollection",
    "StageClassOrInstance",
    "StageCollection",
    "ChainedProcessor",
    "ChainedProcessorInterface",
)



from typing import Any, Optional, Protocol, Self

# extends: 3rd-party interface
from ....support.act_as_factory import ActAsFactoryInterface
from ....support.act_as_list import ActAsListInterface

# uses: bridge interface
from ..bridges.pipeline_interface import TPipeline
from ..bridges.stage_callable import StageClassOrInstance, StageCollection


class PipelineFactoryInterface(
    ActAsFactoryInterface[TPipeline],
    ActAsListInterface[StageClassOrInstance, StageCollection],
    Protocol[TPipeline],
):
    def __init__(
        self,
        stages: Optional[StageCollection] = None,
        pipeline_class: Optional[type[TPipeline]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def add_stage(self, stage: StageClassOrInstance) -> Self: ...

    def with_stages(self, stages: StageCollection) -> Self: ...

    def _get_default_pipeline_class(self) -> Optional[type[TPipeline]]: ...



from abc import ABC
from typing import Any, Optional, Self

# extends: 3rd-party concrete
from ....support.act_as_list import ActAsList

# uses: bridge interfaces
from ..bridges.stage_callable import StageInstance, StageInstanceCollection

# implements: self-interface
from .pipeline_interface_mixin import PipelineInterfaceMixin as ImplementsInterface


class PipelineMixin(
    ActAsList[StageInstance, StageInstanceCollection],
    ImplementsInterface,
    ABC,
):
    stage_instances: StageInstanceCollection

    def __init__(
        self,
        stage_instances: Optional[StageInstanceCollection] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not hasattr(self, "stage_instances"):
            self.stage_instances = tuple()

        if stage_instances:
            self.stage_instances = stage_instances

    def _get_items(self) -> StageInstanceCollection:
        return self.stage_instances

    def _set_items(self, items: StageInstanceCollection) -> Self:
        self.stage_instances = items

        return self

    def _add_item(self, item: StageInstance) -> Self:
        return self.clone({"stage_instances": tuple([*self.stage_instances, item])})

    def pipe(self, stage: StageInstance) -> Self:
        """
        Adds a single stage to the pipeline.
        """
        return self._add_item(stage)



from typing import Protocol

# extends: outside base
from ..._01_with_base import StageCallable as StageCallableBase

# uses: local bridge
from .types import T_in, T_out


class StageCallable(
    StageCallableBase[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass


StageInstance = StageCallable

StageInstanceCollection = tuple[StageInstance, ...]

StageClassOrInstance = StageInstance | type[StageInstance]

StageCollection = tuple[StageClassOrInstance, ...]



from abc import ABC
from typing import Any, Optional, Self

# uses: bridge interface
from ..bridges.stage_callable import StageCollection

# implements: self-interface
from .pipeline_interface_mixin import PipelineInterfaceMixin as ImplementsInterface


class PipelineMixin(
    ImplementsInterface,
    ABC,
):
    stages: StageCollection

    def __init__(
        self,
        stages: Optional[StageCollection] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if not hasattr(self, "stages"):
            self.stages = tuple()

        if stages:
            self.stages = stages

        if self._should_instantiate_stages():
            self._instantiate_stages()

    def _should_instantiate_stages(self) -> bool:
        return len(self._get_items()) == 0 and len(self.stages) > 0

    def _instantiate_stages(self) -> Self:
        instances = tuple(
            stage() if isinstance(stage, type) else stage for stage in self.stages
        )

        return self._set_items(instances)

    def with_stages(self, stages: StageCollection) -> Self:
        """
        Adds a collection of stages to the pipeline.
        """
        cloned = self.clone({"stages": stages, "stage_instances": []})

        return cloned._instantiate_stages()



from ._01_with_base import __all__ as _01_with_base
from ._02_with_pipeline_as_list import __all__ as _02_with_pipeline_as_list
from ._03_with_stage_as_callable import __all__ as _03_with_stage_as_callable
from ._04_with_pipeline_declared_stages import (
    __all__ as _04_with_pipeline_declared_stages,
)
from ._05_with_pipeline_factory import __all__ as _05_with_pipeline_factory
from ._06_with_pipeline_processor import __all__ as _06_with_pipeline_processor
from ._08_with_pipeline_as_stage import __all__ as _08_with_pipeline_as_stage
from ._99_with_pipeline_default_processor import (
    __all__ as _99_with_pipeline_default_processor,
)

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    *_01_with_base,
    *_02_with_pipeline_as_list,
    *_03_with_stage_as_callable,
    *_04_with_pipeline_declared_stages,
    *_05_with_pipeline_factory,
    *_06_with_pipeline_processor,
    *_08_with_pipeline_as_stage,
    *_99_with_pipeline_default_processor,
)



# extends: outside base
from ..plugins._01_with_base import Stage as StageBase

# extends: outside mixins
from ..plugins._03_with_stage_as_callable import (
    StageMixin as _03_WithStageAsCallable,
)

# implements: self-interface
from .stage_interface import StageInterface as ImplementsInterface

# uses: bridge
from .types import T_in, T_out


class Stage(
    _03_WithStageAsCallable[T_in, T_out],
    StageBase,
    ImplementsInterface[T_in, T_out],
):
    pass



from typing import Any, Protocol

# extends: "3rd-party" interface
from ...bridges.processor_interface import ProcessorInterface

# uses: bridge interface
from ...bridges.stage_callable import StageInstanceCollection
from ...bridges.types import T_in, T_out


class ChainedProcessorInterface(
    ProcessorInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    """
    A processor that processes the payload through a series of stages.
    """

    async def process(
        self,
        payload: T_in,
        stages: StageInstanceCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...



from typing import Any, Optional, Self

# extends: 3rd-party concrete
from ....support.act_as_factory import ActAsFactory
from ....support.act_as_list import ActAsList
from ..bridges.pipeline_interface import TPipeline

# uses: bridge interfaces
from ..bridges.stage_callable import StageClassOrInstance, StageCollection

# implements: self-interface
from .pipeline_factory_interface import PipelineFactoryInterface as ImplementsInterface


class PipelineFactory(
    ActAsFactory[TPipeline],
    ActAsList[StageClassOrInstance, StageCollection],
    ImplementsInterface[TPipeline],
):
    stages: StageCollection  # ActAsList
    pipeline_class: Optional[type[TPipeline]]

    def __init__(
        self,
        stages: Optional[StageCollection] = None,
        pipeline_class: Optional[type[TPipeline]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        if not hasattr(self, "stages"):
            self.stages = tuple()

        if not hasattr(self, "pipeline_class"):
            self.pipeline_class = self._get_default_pipeline_class()

        if stages:
            self.stages = stages

        if pipeline_class:
            self.pipeline_class = pipeline_class

    # ActAsList
    def _get_items(self) -> StageCollection:
        return self.stages

    # ActAsList
    def _set_items(self, items: StageCollection) -> Self:
        self.stages = items

        return self

    # ActAsList
    def _add_item(self, item: StageClassOrInstance) -> Self:
        return self._set_items(self.stages + (item,))

    # ActAsFactory
    def _get_model_class(self) -> type[TPipeline]:
        if self.pipeline_class is None:
            raise ValueError("Pipeline class not set in factory.")

        return self.pipeline_class

    # ActAsFactory
    def _definition(self) -> dict[str, Any]:
        return {"stages": self._get_items()}

    # public API: factory configuration
    def add_stage(self, stage: StageClassOrInstance) -> Self:
        """
        Adds a single stage to the pipeline.
        """
        return self._add_item(stage)

    # public API: factory configuration
    def with_stages(self, stages: StageCollection) -> Self:
        """
        Adds a collection of stages to the pipeline.
        """
        return self._set_items(stages)

    def _get_default_pipeline_class(self) -> Optional[type[TPipeline]]:
        return None



from typing import Any

# extends: 3rd-party concrete
from ....support.clonable import Clonable

# implements: self-interface
from .pipeline_interface import PipelineInterface as ImplementsInterface


class Pipeline(
    Clonable,
    ImplementsInterface,
):
    def __init__(self, *args: Any, **kwds: Any) -> None:
        pass



from typing import Any, Optional, Protocol, Self

# uses: local base
from ..bases.processor_interface import ProcessorInterface

# extends: self-bridge
from ..bridges.pipeline_interface import PipelineInterface

# uses: bridge interface
from ..bridges.types import T_in, T_out


class PipelineInterfaceMixin(
    PipelineInterface,
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        processor_class: Optional[type[ProcessorInterface]] = None,
        processor_instance: Optional[ProcessorInterface] = None,
        processor: Optional[type[ProcessorInterface] | ProcessorInterface] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def with_processor(
        self, processor: type[ProcessorInterface] | ProcessorInterface
    ) -> Self: ...

    def with_processor_instance(
        self, processor_instance: ProcessorInterface
    ) -> Self: ...

    def with_processor_class(
        self, processor_class: type[ProcessorInterface]
    ) -> Self: ...

    def _make_processor(self) -> ProcessorInterface: ...

    async def process(self, payload: T_in, *args: Any, **kwds: Any) -> T_out: ...

    def get_processor_instance(self) -> Optional[ProcessorInterface]: ...

    def get_processor_class(self) -> Optional[type[ProcessorInterface]]: ...

    def _get_default_processor_class(self) -> Optional[type[ProcessorInterface]]: ...



from typing import Protocol

# extends: outside base
from ..plugins._01_with_base import StageCallable as StageCallableBase
from ..plugins._01_with_base import (
    StageClassOrInstance,
    StageCollection,
    StageInstance,
    StageInstanceCollection,
)

# uses: bridge interface
from .types import T_in, T_out


class StageCallable(
    StageCallableBase[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass


__all__ = (
    "StageCallable",
    "StageInstance",
    "StageInstanceCollection",
    "StageClassOrInstance",
    "StageCollection",
)



from abc import ABC
from typing import Any

# uses: bridge interface
from ..bridges.types import T_in, T_out

# implements: self-interface
from .pipeline_interface_mixin import PipelineInterfaceMixin as ImplementsInterface


class PipelineMixin(
    ImplementsInterface[T_in, T_out],
    ABC,
):
    async def __call__(
        self,
        payload: T_in,
        /,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Processes the payload through the pipeline.
        """
        return await self.process(payload, *args, **kwds)



from abc import abstractmethod
from typing import Any, Protocol, TypeVar

# extends: 3rd-party interface
from ....support.clonable import ClonableInterface
from ....support.has_call_async import HasCallAsyncInterface

# uses: bridge interface
from ..bridges.stage_callable import StageInstance, StageInstanceCollection
from ..bridges.types import T_in, T_out


class ProcessorInterface(
    HasCallAsyncInterface,
    ClonableInterface,
    Protocol[T_in, T_out],
):
    def __init__(self, *args: Any, **kwds: Any) -> None: ...

    @abstractmethod
    async def process(
        self,
        payload: T_in,
        stages: StageInstanceCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    # HasCallAsync: include callable type and `payload` in the signature
    async def _call(
        self,
        callable: StageInstance,
        payload: T_in,
        *args: Any,
        **kwds: Any,
    ) -> Any: ...


TProcessor = TypeVar("TProcessor", bound=ProcessorInterface, infer_variance=True)



from abc import ABC, abstractmethod
from typing import Any

# uses: bridge
from ..bridges.types import T_in, T_out

# implements: self-interface
from .stage_interface_mixin import StageInterfaceMixin as ImplementsInterface


class StageMixin(
    ImplementsInterface[T_in, T_out],
    ABC,
):
    @abstractmethod
    async def __call__(
        self,
        payload: T_in,
        /,  # Make 'payload' a positional-only parameter
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Runs the stage.
        """
        pass



from typing import Protocol

# extends: outside base
from ..._05_with_pipeline_factory import (
    PipelineFactoryInterface as PipelineFactoryBaseInterface,
)

# extends: outside mixins
from ..._06_with_pipeline_processor import (
    PipelineFactoryInterfaceMixin as _06_WithPipelineProcessorInterface,
)

# uses: bridge interface
from .pipeline_interface import PipelineInterface
from .types import T_in, T_out


class PipelineFactoryInterface(
    _06_WithPipelineProcessorInterface[T_in, T_out],
    PipelineFactoryBaseInterface[PipelineInterface[T_in, T_out]],
    Protocol[T_in, T_out],
):
    pass



# Re-exporting symbols
from .final import ChainedProcessor as ChainedProcessor
from .final import Pipeline as Pipeline
from .final import PipelineFactory as PipelineFactory
from .final import PipelineFactoryInterface as PipelineFactoryInterface
from .final import PipelineInterface as PipelineInterface
from .final import Processor as Processor
from .final import ProcessorInterface as ProcessorInterface
from .final import Stage as Stage
from .final import StageCallable as StageCallable
from .final import StageClassOrInstance as StageClassOrInstance
from .final import StageCollection as StageCollection
from .final import StageInstance as StageInstance
from .final import StageInstanceCollection as StageInstanceCollection
from .final import StageInterface as StageInterface
from .final import T_in as T_in
from .final import T_out as T_out
from .final import __all__ as _final_all

# pyright: reportUnsupportedDunderAll=false
__all__ = (*_final_all,)



"""Library's public types"""

# Re-exporting symbols
from _api.core import StageClassOrInstance as StageClassOrInstance
from _api.core import StageCollection as StageCollection
from _api.core import StageInstance as StageInstance
from _api.core import StageInstanceCollection as StageInstanceCollection
from _api.core import T_in as T_in
from _api.core import T_out as T_out

# pyright: reportUnsupportedDunderAll=false
__all__ = (
    "StageClassOrInstance",
    "StageCollection",
    "StageInstance",
    "StageInstanceCollection",
    "T_in",
    "T_out",
)



from typing import Any, cast

# extends: "3rd-party" concrete
from ...bridges.processor import Processor

# uses: bridge interface
from ...bridges.stage_callable import StageInstanceCollection
from ...bridges.types import T_in, T_out

# implements: self-interface
from .chained_processor_interface import (
    ChainedProcessorInterface as ImplementsInterface,
)


class ChainedProcessor(
    Processor[T_in, T_out],
    ImplementsInterface[T_in, T_out],
):
    async def process(
        self,
        payload: T_in,
        stages: StageInstanceCollection,
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        """
        Process the given payload through the provided stages.

        Args:
            payload (T_in): The input payload to process.
            stages (StageInstanceCollection): The collection of stages to process the payload through.
            *args (Any): Additional positional arguments.
            **kwds (Any): Additional keyword arguments.

        Returns:
            T_out: The processed output.
        """
        for stage in stages:
            payload = await self._call(callable=stage, payload=payload, *args, **kwds)

        return cast(T_out, payload)



from typing import Protocol

# extends: outside base
from ..plugins._01_with_base import StageInterface as StageBaseInterface

# extends: outside mixins
from ..plugins._03_with_stage_as_callable import (
    StageInterfaceMixin as _03_WithStageAsCallableInterface,
)

# uses: bridge interface
from .types import T_in, T_out


class StageInterface(
    _03_WithStageAsCallableInterface[T_in, T_out],
    StageBaseInterface,
    Protocol[T_in, T_out],
):
    pass



from .bases.processors.chained_processor import ChainedProcessor
from .bases.processors.chained_processor_interface import ChainedProcessorInterface
from .mixins.pipeline_factory_interface_mixin import PipelineFactoryInterfaceMixin
from .mixins.pipeline_factory_mixin import PipelineFactoryMixin
from .mixins.pipeline_interface_mixin import PipelineInterfaceMixin
from .mixins.pipeline_mixin import PipelineMixin

__all__ = (
    "ChainedProcessorInterface",
    "ChainedProcessor",
    "PipelineMixin",
    "PipelineInterfaceMixin",
    "PipelineFactoryMixin",
    "PipelineFactoryInterfaceMixin",
)



from .act_as_list import ActAsList
from .act_as_list_interface import ActAsListInterface
from .types import TCollection, TItem

__all__ = (
    # main
    "ActAsList",
    "ActAsListInterface",
    # types
    "TItem",
    "TCollection",
)



from typing import Protocol, Self

from .types import TItem, TCollection


class ActAsListInterface(
    Protocol[TItem, TCollection],
):
    def _get_items(self) -> TCollection: ...

    def _set_items(self, items: TCollection) -> Self: ...

    def _add_item(self, item: TItem) -> Self: ...



from typing import Protocol, TypeVar

# extends: outside base
from ..._01_with_base import PipelineInterface as PipelineBaseInterface


class PipelineInterface(
    PipelineBaseInterface,
    Protocol,
):
    pass


TPipeline = TypeVar("TPipeline", bound=PipelineInterface, infer_variance=True)



from abc import ABC

# uses: base concrete
from ..bases.processors.chained_processor import ChainedProcessor

# extends: self-bridge concrete
from ..bridges.pipeline import Pipeline

# uses: bridge interface
from ..bridges.types import T_in, T_out

# implements: self-interface
from .pipeline_interface_mixin import PipelineInterfaceMixin as ImplementsInterface


class PipelineMixin(
    Pipeline[T_in, T_out],
    ImplementsInterface[T_in, T_out],
    ABC,
):
    def _get_default_processor_class(self) -> type[ChainedProcessor[T_in, T_out]]:
        return ChainedProcessor[T_in, T_out]



from typing import Generic

# extends: outside base
from ..._01_with_base import Pipeline as PipelineBase

# extends: outside mixins
from ..._02_with_pipeline_as_list import (
    PipelineMixin as _02_WithPipelineAsList,
)
from ..._04_with_pipeline_declared_stages import (
    PipelineMixin as _04_WithPipelineDeclaredStages,
)
from ..._06_with_pipeline_processor import (
    PipelineMixin as _06_WithPipelineProcessor,
)
from ..._08_with_pipeline_as_stage import (
    PipelineMixin as _08_WithPipelineAsStage,
)

# implements: self-interface
from .pipeline_interface import PipelineInterface as ImplementsInterface

# uses: bridge interface
from .types import T_in, T_out


class Pipeline(
    _08_WithPipelineAsStage[T_in, T_out],
    _06_WithPipelineProcessor[T_in, T_out],
    _04_WithPipelineDeclaredStages,
    _02_WithPipelineAsList,
    PipelineBase,
    ImplementsInterface[T_in, T_out],
    Generic[T_in, T_out],
):
    pass
